
program Bachelor;

{* Bachelor is a freeware program for educational or scientific purposes
 * in propositional logic.
 * Author:Hashim Habiballa, Department of Informatics, University of Ostrava,
 *         Czech republic  (bachelor project for 1997 , (*) 1995 - 6 )
 * Details: Syntactical trees,full functional set of operators,normal forms,
 *          mimimal NF, decisions about realizability and validity         }

{$M 32768,8192,655000}
{$X+,S-}

uses Dos, Objects, Drivers, Memory, Views, Menus, Dialogs,
  StdDlg, MsgBox, App, Editors, HelpFile, Deal;

const
  HeapSize = 8 * (1024 div 16);
  cmOpen       = 100;
  cmNew        = 101;
  cmChangeDir  = 102;
  cmDosShell   = 103;
  cmCalculator = 104;
  cmShowClip   = 105;
  cmRetrieve   = 106;
  cmNewProt    = 108;
  cmRefresh    = 109;
  cmFormat     = 110;
  cmDNF        = 114;
  cmCNF        = 115;
  cmCon        = 116;
  cmDis        = 117;
  cmConDis     = 118;
  cmImp        = 119;
  cmCompile    = 120;
  cmDecompile  = 121;
  cmDecision   = 122;
  cmEval       = 123;
  cmMinimalDNF = 124;
  cmMinimalCNF = 125;
  cmContents   = 126;
  cmAbout      = 127;
  cmSystem     = 128;
  cmPL         = 129;
  edOK         = 20000;
  edNothing    = 20001;
  edillchar    = 20002;
  edmissoper   = 20003;
  edmissexp    = 20004;
  edmissbra    = 20005;
  edunexpbra   = 20006;
  edinsertfail = 20007;
  ending=#0; {ending character}
  andchar:char='&';
  orchar:char='|';
  TIgnored:set of char=[#1..#32]; {ignored characters}
  TLegal:set of char=['a'..'z','A'..'Z','0','1',ending,'&','|','~','>',
                          '=','(',')' ]; {legal characters}

type
  PEditorApp = ^TEditorApp;
  TEditorApp = object(TApplication)
    constructor Init;
    procedure HandleEvent(var Event: TEvent); virtual;
    procedure InitMenuBar; virtual;
    procedure InitStatusLine; virtual;
    procedure OutOfMemory; virtual;
    procedure GetEvent(var Event:TEvent);virtual;
    function GetPalette:PPalette;virtual;
  end;

     PSTreeNode=^TSTreeNode;   {tree node}
     TSTreeNode=record
      character:char;          {symbol : operator,proposition,constant}
      prior:byte;              {priority of symbol}
      neg:boolean;             {setting of negation}
      left:PSTreeNode;         {pointer to left subtree}
      right:PSTreeNode;                 {to right subtree}
     end;
     

     PSStackNode=^TSStackNode; { stack node containg tree node} 
     TSStackNode=record        {during compilation}
      node:PSTreeNode;         {pointer to tree node}
      next:PSStackNode;        {pointer to next stack node}
     end;
     ErrorType=(OK,illchar,missoper,missexp,missbra,unexpbra,endcomp,
                insfail,mem);
       {error types = OK , illegal character , missing operator , expression,
                      bracket , unexpected bracket ,end of compilation }
     PSStack=PSStackNode;

  PProtocol = ^TProtocol;

  PPLFormula = ^TPLFormula;
  TPLFormula = object
    Stree: PSTreeNode;
    Protocol : PProtocol;
    constructor Init;
    destructor Done;virtual;
    procedure DisposeTree;
    procedure CopyTree(var p2:PSTreeNode);
    procedure ConstEval;
    procedure ToImp;
    procedure ToDis;
    procedure ToCon;
    procedure ToConDis;
    procedure ToCNF;
    procedure ToDNF;
    procedure RDisposeTree(var tr:PSTreeNode);
    procedure RCopyTree(var p1,p2:PSTreeNode);
    procedure RConstEval(var tr:PSTreeNode);
    procedure RToImp(var tr:PSTreeNode);
    procedure RToDis(var tr:PSTreeNode);
    procedure RToCon(var tr:PSTreeNode);
    procedure RToConDis(var tr:PSTreeNode);
    procedure RToCNF(var tr:PSTreeNode);
    procedure RToDNF(var tr:PSTreeNode);
    procedure FindOper(var p1:PSTreeNode;var pch:char);
    procedure ChangeOper(p:PStreeNode;ch:char);
    procedure EvlConst1(var p:PSTreeNode);
    procedure Implication(var p:PStreeNode);
    procedure Disjunction(var p:PStreeNode);
    procedure Conjunction(var p:PSTreeNode);
    procedure Distribution(var p1,p2:PSTreeNode);
    procedure CopyNode(var p1,p2 : PSTreeNode);
    procedure ShCopyNode(var p1,p2 : PSTreeNode);
    procedure Exchange(var p1,p2 : PSTreenode);
    procedure SortElm(var tr: PSTreeNode);
    procedure ReduceElm(var t:PSTreeNode);
    procedure IdCom(var p1,p2,master:PSTreeNode);
    procedure Thread;
    procedure UnThread;
    procedure Minimize;
    procedure AbsExt(var p1,p2:PSTreeNode;var o:boolean;pchar:char);
  end;

  PFormat = ^TFormat;
  TFormat = record
   line   : integer;
   space : boolean;
   results: boolean;
   manner : word;
  end;

  TProtocol =  object(TEditWindow)
    Formula : PPLFormula;
    Format  : PFormat;
    Error   : ErrorType;
    constructor Init(var Bounds: TRect;
      FileName: FNameStr; ANumber: Integer);
    destructor Done;virtual;
    procedure HandleEvent(var Event:TEvent);virtual;
    procedure XComp;
    procedure XDecomp;
    procedure InsertCharacter(x:char);
    procedure InsertLine;
    procedure InsertString(s:string);
    procedure InsertFail;
    procedure Dive(var node:PSTreeNode);
    procedure DoDialog;
    procedure Decision;
  end;
  PDialogData = ^DialogData;
  DialogData = record
    RBData1: Word;
    RBData2: Word;
    RBData3: Word;
    InputLineData: string[3];
  end;

  PDemoDialog = ^TDemoDialog;
  TDemoDialog = object(TDialog)

  end;


var
  EditorApp: TEditorApp;
  ClipWindow: PEditWindow;

function ExecDialog(P: PDialog; Data: Pointer): Word;
var
  Result: Word;
begin
  Result := cmCancel;
  P := PDialog(Application^.ValidView(P));
  if P <> nil then
  begin
    if Data <> nil then P^.SetData(Data^);
    Result := DeskTop^.ExecView(P);
    if (Result <> cmCancel) and (Data <> nil) then P^.GetData(Data^);
    Dispose(P, Done);
  end;
  ExecDialog := Result;
end;


function DoEditDialog(Dialog: Integer; Info: Pointer): Word; far;
var
  R: TRect;
  T: TPoint;pom:string[10];
begin
  case Dialog of
    edOutOfMemory:
      DoEditDialog := MessageBox('Not enough memory for this operation.',
        nil, mfError + mfOkButton);
    edReadError:
      DoEditDialog := MessageBox('Error reading file %s.',
        @Info, mfError + mfOkButton);
    edWriteError:
      DoEditDialog := MessageBox('Error writing file %s.',
        @Info, mfError + mfOkButton);
    edCreateError:
      DoEditDialog := MessageBox('Error creating file %s.',
        @Info, mfError + mfOkButton);
    edSaveModify:
      DoEditDialog := MessageBox('%s has been modified. Save?',
        @Info, mfInformation + mfYesNoCancel);
    edSaveUntitled:
      DoEditDialog := MessageBox('Save untitled file?',
        nil, mfInformation + mfYesNoCancel);
    edSaveAs:
      DoEditDialog := ExecDialog(New(PFileDialog, Init('*.*',
        'Save file as', '~N~ame', fdOkButton, 101)), Info);
    edOK:
      DoEditDialog := MessageBox('Compilation successful.',
        nil, mfInformation + mfOKButton);
    edNothing:
          DoEditDialog := MessageBox('No selected text.',
        nil, mfError + mfOKButton);
    edillchar:
          DoEditDialog := MessageBox('Illegal character.',
        nil, mfError + mfOKButton);
    edmissoper:
          DoEditDialog := MessageBox('Missing operator.',
        nil, mfError + mfOKButton);
    edmissexp:
          DoEditDialog := MessageBox('Missing expression.',
        nil, mfError + mfOKButton);
    edmissbra:
          DoEditDialog := MessageBox('Missing right bracket.',
        nil, mfError + mfOKButton);
    edunexpbra:
          DoEditDialog := MessageBox('Unexpected right bracket.',
        nil, mfError + mfOKButton);
    edinsertfail:
          DoEditDialog := MessageBox('Breaking down during inserting a text.',
        nil, mfError + mfOKButton);
  end;
end;

function OpenEditor(FileName: FNameStr; Visible: Boolean): PEditWindow;
var
  P: PView;
  R: TRect;
begin
  DeskTop^.GetExtent(R);
  P := Application^.ValidView(New(PEditWindow,
    Init(R, FileName, wnNoNumber)));
  if not Visible then P^.Hide;
  P^.DisableCommands([cmEval,cmFormat,
    cmMinimalCNF,cmMinimalDNF,cmCNF,cmDecision,cmDNF,cmImp,cmDecompile,
    cmCompile,cmConDis,cmCon,cmDis]);
  DeskTop^.Insert(P);
  OpenEditor := PEditWindow(P);
end;

function OpenProtocol(FileName: FNameStr; Visible: Boolean): PProtocol;
var
  P: PView;
  R: TRect;
begin
  DeskTop^.GetExtent(R);
  P := Application^.ValidView(New(PProtocol,
    Init(R, FileName, wnNoNumber)));
  if not Visible then P^.Hide;
  P^.EnableCommands([cmEval,cmFormat,
    cmMinimalCNF,cmMinimalDNF,cmCNF,cmDecision,cmDNF,cmImp,cmDecompile,
    cmCompile,cmConDis,cmCon,cmDis]);
  DeskTop^.Insert(P);
  OpenProtocol := PProtocol(P);
end;

procedure TProtocol.DoDialog;
var
  DData:PDialogData;
  Bruce: PView;
  Dialog: PDemoDialog;
  R: TRect;
  C: Word;pom:string;i:integer;
begin
 str(Format^.line,pom);
 DData^.InputLineData:=pom;
 if Format^.space  then DData^.RBData1:=0 else DData^.RBData1:=1;
 if Format^.results then DData^.RBData2:=0 else DData^.RBData2:=1;
 DData^.RBData3:=Format^.manner;
 repeat
  R.Assign(10, 6, 70, 18);
  Dialog := New(PDemoDialog, Init(R, 'Protocol Format'));
  with Dialog^ do
  begin
    R.Assign(3, 3, 25 , 5);
    Bruce := New(PRadioButtons, Init(R,
      NewSItem('~S~paces',
      NewSItem('~N~o Spaces',
      nil))
    ));
    Insert(Bruce);
    R.Assign(3, 6, 25, 8);
    Bruce := New(PRadioButtons, Init(R,
      NewSItem('~R~esults ',
      NewSItem('~W~ithout Results ',
      nil))
    ));
    Insert(Bruce);
    R.Assign(2, 2, 15, 3);
    Insert(New(PLabel, Init(R, 'Characters', Bruce)));
    R.Assign(2, 5, 15, 6);
    Insert(New(PLabel, Init(R, 'Operations', Bruce)));
    Insert(Bruce);
    R.Assign(27, 3, 55, 8);
    Bruce := New(PRadioButtons, Init(R,
      NewSItem('~C~omment + Evaluations ',
      NewSItem('~S~tep by step + Comment',
      NewSItem('~O~nly step by step',
      NewSItem('~M~acrosteps',
      NewSItem('~N~othing',
      nil)))))
    ));
    Insert(Bruce);
    R.Assign(27, 2, 50, 3);
    Insert(New(PLabel, Init(R, 'Details', Bruce)));
    R.Assign(17, 9, 22, 10);
    Bruce := New(PInputLine, Init(R, 3));
    Insert(Bruce);
    R.Assign(2, 9, 15, 10);
    Insert(New(PLabel, Init(R, 'Line width', Bruce)));
    R.Assign(27,9,37,11 );
    Insert(New(PButton, Init(R, '~O~k', cmOK, bfDefault)));
    R.Assign(42,9,52,11);
    Insert(New(PButton, Init(R, 'Cancel', cmCancel, bfNormal)));
  end;
  C := ExecDialog(Dialog,DData);
  pom:=DData^.InputLineData;
  val(pom,Format^.line,i);
  if (Format^.line <= 5)or(Format^.line>250)
   then MessageBox('Current line range between 6 and 250.'
      ,nil,mfError+mfOKButton);
 until (Format^.line>5)and(Format^.line<=250);
 Format^.space:=(DData^.RBData1=0);
 Format^.results:=(DData^.RBData2=0);
 Format^.manner:=DData^.RBData3;
end;


procedure TProtocol.XComp;{analysis of a formula and generating of a tree
                     by recursion}
                    { Exp1 ::= Exp2  ['=' Exp2];   [] - means iteration
                      Exp2 ::= Exp3  ['>' Exp3];   ' '- means terminal char.
                      Exp3 ::= Exp4  ['|' Exp4];
                      Exp4 ::= ICE   ['&' ICE];
                      ICE ::= '~' ICE | '(' Exp1 ')' | '0' | '1' |
                              | 'A'| .. | 'Z' | 'a' | .. | 'z' ;

                      '0' - false , '1' - true , 'A'..'Z','a'..'z' - atoms ,
                      '~' - negation , '=' - equivalence , '>' - implication ,
                      '|' - disjunction , '&' - conjuction  }
  var pom,garbage:PSStackNode;
      HeapPoint:pointer;
      VPSStack:PSStack;
      ch:char;
      i:integer;
      pos:word;

  procedure StackInit;     {Initilizing of stack}
   begin
    mark(HeapPoint);
    VPSStack:=nil;
   end;

   procedure Put(var chp:char;pr:byte);  {putting of stack node}
    begin
    if not(LowMemory) then
     begin
      new(pom);
      pom^.next:=VPSStack;
      VPSStack:=pom;
      new(pom^.node);
      VPSStack^.node^.character:=chp;
      pom^.node^.left:=nil;
      pom^.node^.right:=nil;
      pom^.node^.prior:=pr;
      pom^.node^.neg:=false;
     end
     else error:=mem;
    end;

   function Cut:pointer;     {cutting of stack node}
    begin
     if VPSStack<>nil then
      begin
       Cut:=VPSStack;
       VPSStack:=VPSStack^.next;
      end;
    end;

   procedure SInsert(var p:PSStackNode);  {inserting of existing stack node}
    begin
     p^.next:=VPSStack;
     VPSStack:=p;
    end;
 procedure Action(chr:char;pr:byte);   {putting of operator to stack}
 begin
   Put(chr,pr);
   pom:=Cut;
   pom^.node^.right:=VPSStack^.node;
   garbage:=Cut;
   dispose(garbage);
   pom^.node^.left:=VPSStack^.node;
   garbage:=Cut;
   dispose(garbage);
   SInsert(pom);
 end;

 procedure GetChar;  {reading of character}
  begin
    if Error=OK then
      begin
        repeat
          if pos>=Editor^.SelEnd then
            ch:=ending
          else
            begin
              ch:=Editor^.BufChar(pos);
              inc(pos);
            end;
        until not(ch in TIgnored);
        if (not(ch in TLegal))and(error=OK) then Error:=illchar;
      end;
  end;

 procedure Exp1;forward;
 procedure ICE;
 begin
 if error=OK then
  begin
   case ch of
    '~':begin
          Getchar;
          ICE;
          VPSStack^.node^.neg:=not(VPSStack^.node^.neg);
        end;
    'a'..'z','A'..'Z','0'..'1':
        begin
          Put(ch,6);
          Getchar;
        end;
    '(':begin
          Getchar;
          Exp1;
          if (ch<>')')and(error=OK) then Error:=missbra;
          Getchar;
        end
    else Error:=missexp;
    end;
  end;
 end;
 procedure Exp4;
 begin
  if Error=OK then
    begin
      ICE;
      while (ch='&')and(error=OK) do
        begin
         Getchar;
         ICE;
         if error=OK then Action('&',4);
        end;
    end;
 end;

 procedure Exp3;
 begin
  if Error=OK then
    begin
      Exp4;
      while (ch='|')and(error=OK) do
        begin
         Getchar;
         Exp4;

         if error=OK then Action('|',3);

        end;
    end;
 end;

 procedure Exp2;
 begin
  if Error=OK then
    begin
      Exp3;
      while (ch='>')and(error=OK) do
        begin
         Getchar;
         Exp3;
         if error=OK then Action('>',2);
        end;
    end;
 end;

 procedure Exp1;
 begin
  if Error=OK then
    begin
      Exp2;
      while (ch='=')and(error=OK) do
        begin
         Getchar;
         Exp2;
         if error=OK then Action('=',1);
        end;
    end;
 end;


 begin
  StackInit;
  pos:=Editor^.SelStart;
  Error:=OK;
  Getchar;
  if Error=OK then Exp1;
  if (ch=')')and(error=OK) then Error:=unexpbra;
  if (ch=ending)and(error=OK) then Error:=endcomp;
  if (ch<>ending)and(error=OK) then Error:=missoper;
  if Error=endcomp then
   begin
      Error:=OK;
      Formula^.Stree:=VPSStack^.Node;
      dispose(VPSStack);
      Editor^.SetSelect(pos,pos,true);
      InsertLine;
   end
  else
     begin
      release(HeapPoint);
      Editor^.SetSelect(pos - 1,pos - 1,true);
      InsertString(' ');
      Editor^.SetSelect(pos - 1,pos - 1,true);
     end;
 case error of
  illchar: i:=DoEditDialog(edillchar,nil);
  missoper: i:=DoEditDialog(edmissoper,nil);
  missbra: i:=DoEditDialog(edmissbra,nil);
  missexp: i:=DoEditDialog(edmissexp,nil);
  unexpbra: i:=DoEditDialog(edunexpbra,nil);
  mem : i:=DoEditDialog(edOutOfMemory,nil);
 end;
  
 end;      {end of XCOMP}

 procedure TProtocol.Decision;
 begin
   if Formula<>nil then
     begin
       Formula^.ToDNF;
       Formula^.Minimize;
       case Formula^.Stree^.Character of
       '0': begin
              InsertString('This formula is not consistent.');
              InsertLine;
            end;
       '1': begin
              InsertString('This formula is valid.');
              InsertLine;
            end;
       else
         begin
           InsertString('This formula is consistent.Its models '+
           'can be expressed as DNF:');
           InsertLine;
           XDecomp;
         end;
       end;
     end;
 end;

 procedure TPLFormula.RDisposeTree;
 begin

       if tr^.prior<>6 then
         begin
           if tr^.left<>nil then RDisposeTree(tr^.left);
           if tr^.right<>nil then RDisposeTree(tr^.right);
         end;
       dispose(tr);
 end;

 procedure TPLFormula.DisposeTree;
 begin
   if Stree<>nil then RDisposeTree(Stree);
 end;

 procedure TPLFormula.CopyNode;
 begin
   if p1<> nil then
   begin
     new(p2);
     p2^.character:=p1^.character;
     p2^.left:=p1^.left;
     p2^.right:=p1^.right;
     p2^.neg:=p1^.neg;
     p2^.prior:=p1^.prior;
   end;
 end;

 procedure TPLFormula.ShCopyNode;
 begin
   if (p1<> nil)and(p2<>nil) then
   begin
     p2^.character:=p1^.character;
     p2^.left:=p1^.left;
     p2^.right:=p1^.right;
     p2^.neg:=p1^.neg;
     p2^.prior:=p1^.prior;
   end;
 end;


 
 procedure TPLFormula.RCopyTree;
 begin

       CopyNode(p1,p2);
       if p1^.prior<>6 then
         begin
           if p1^.left<>nil then RCopyTree(p1^.left,p2^.left);
           if p1^.right<>nil then RCopyTree(p1^.right,p2^.right);
         end;
 end;

 procedure TPLFormula.CopyTree;
 begin
   if Stree<>nil then RCopyTree(Stree,p2);
 end;

 procedure TPLFormula.FindOper;
 var pompt2,pompt3:PSTreeNode;
   procedure Find1(var p:PSTreeNode);
   begin
     if pompt2=nil then
     begin
       if p^.character=pch then
         begin
           pompt2:=p;p:=pompt3;
         end;
       if p^.prior<>6 then
         begin
           if p^.left<>nil then Find1(p^.left);
           if p^.right<>nil then Find1(p^.right);
         end;
     end;
   end;

    begin
      pompt2:=nil;pompt3:=p1^.right;
      if p1^.left<>nil then Find1(p1^.left);
      if pompt2<>nil then
        begin
          p1^.right:=p1^.left;p1^.left:=pompt2;
        end
      else
        begin
          pompt3:=p1^.left;
          if p1^.right<>nil then Find1(p1^.right);
          if pompt2<>nil then p1^.left:=pompt2;
        end;

    end;


 procedure TPLFormula.ChangeOper;
 begin
   if p<>nil then
     begin
       p^.character:=ch;
       case ch of
        '=': p^.prior:=1;
        '>': p^.prior:=2;
        '|': p^.prior:=3;
        '&': p^.prior:=4;
       end;
     end;
  end;

  procedure TProtocol.InsertFail;
  var i:integer;
  begin
    i:=DoEditDialog(edinsertfail,nil);
    error:=insfail;
  end;

  procedure TProtocol.InsertLine;
  var chpoint:^char;
  begin
  if error<>insfail then
    begin
      new(chpoint);
      chpoint^:=#13;
      if not(Editor^.InsertText(chpoint,1,false)) then InsertFail;
      chpoint^:=#10;
      if not(Editor^.InsertText(chpoint,1,false)) then InsertFail;
      dispose(chpoint);
     
     end;
  end;

  procedure TProtocol.InsertString;
  var i:integer;
  begin
    for i:=1 to length(s) do
      begin
       Insertcharacter(s[i]);
       if Format^.Space then Insertcharacter(' ');
      end;
  end;

  procedure TProtocol.InsertCharacter;
  var chpoint:^char;
  begin
    if error<>insfail then
      begin
        new(chpoint);
        if Editor^.CurPos.X >= Format^.line  then InsertLine;
        chpoint^:=x;
        if not(Editor^.InsertText(chpoint,1,false)) then InsertFail;
        dispose(chpoint);
      end;
  end;



   procedure TProtocol.Dive;
   begin
     
       
         if node^.neg=true then
           begin
             InsertString('~');
             if node^.prior<>6 then   {respecting literals with negations}
               begin
                 InsertString('(');
               end;
           end;
         if node^.prior <> 6 then
           if (node^.prior>node^.left^.prior)
               and(not(node^.left^.neg)) then
              begin
                InsertString('(');
                Dive(node^.left);     {respecting priority violation}
                InsertString(')');
              end
           else Dive(node^.left);
         InsertString(node^.character);
         if node^.prior <> 6 then
           if ((node^.prior>node^.right^.prior)or
              ((node^.prior=2)and(node^.right^.prior=2)))
               and(not(node^.right^.neg)) then
              begin
                InsertString('(');
                Dive(node^.right);
                InsertString(')');
              end
           else Dive(node^.right);
         if (node^.prior<>6)and(node^.neg=true) then
           begin
             InsertString(')');
           end;

    
  end;

  procedure TProtocol.XDecomp;
   begin
     if Formula^.Stree<>nil then
      begin
       Dive(Formula^.Stree);
       InsertLine;
      end;
   end;



 { Propositional logic laws }



 procedure TPLFormula.EvlConst1;
 {procedure for evaluating constants of one node}
 var left:boolean;pchar:char;
     pompt1,pompt2:PSTreeNode;
 begin
 if (p<>nil)and(p^.prior<>6) then
 begin
   pompt1:=nil;pompt2:=nil;
   
   if (p^.left^.character in ['0','1'])and(p^.left<>nil) then
     begin
       pompt1:=p^.right;
       pompt2:=p^.left;
       left:=true;
     end
   else if (p^.right^.character in ['0','1'])and(p^.right<>nil) then
          begin
            pompt1:=p^.left;
            pompt2:=p^.right;
            left:=false;
          end;
   if pompt1<>nil then
     begin
       if pompt2^.neg then
         begin                        {removes negs from constants}
           pompt2^.neg:=false;
           if pompt2^.character='0' then pompt2^.character:='1'
           else pompt2^.character:='0';
         end;
     pchar:=p^.character;
     dispose(p);
     case pchar of
      '=':begin
            if pompt2^.character='0' then pompt1^.neg:=not(pompt1^.neg);
            p:=pompt1;RdisposeTree(pompt2);
          end;
                { ( A = 1 ) = A
                  ( A = 0 ) = ~A }

      '>':begin
            if pompt2^.character='0' then
              if left then
                 begin
                   p:=pompt2;p^.character:='1';RdisposeTree(pompt1);
                 end
              else begin
                     p:=pompt1;p^.neg:=not(p^.neg);RdisposeTree(pompt2);
                   end
            else
            if pompt2^.character='1' then
              if left then
                begin
                  p:=pompt1;RdisposeTree(pompt2);
                end
              else begin
                     p:=pompt2;RdisposeTree(pompt1);
                   end;
           end;
              { ( A > 1 ) = 1
                ( A > 0 ) = ~A
                ( 1 > A ) = A
                ( 0 > A ) = 1 }

       '|':if pompt2^.character='0' then
                begin
                  p:=pompt1;RdisposeTree(pompt2);
                end
           else begin
                  p:=pompt2;RdisposeTree(pompt1);
                end;
           { ( A | 1 ) = 1
             ( A | 0 ) = A }

        '&':if pompt2^.character='0' then
                begin
                  p:=pompt2;RdisposeTree(pompt1);
                end
           else begin
                  p:=pompt1;RdisposeTree(pompt2);
                end;
             { ( A & 1 ) = A
               ( A & 0 ) = 0 }

      end;

     if (p^.character='0')and(p^.neg) then
        begin
          p^.character:='1';p^.neg:=false;   {removes negations from const.}
        end
     else if (p^.character='1')and(p^.neg) then
        begin
          p^.character:='0';p^.neg:=false;
        end;
     if Protocol<>nil then
       with Protocol^ do
         
           if Format^.manner = 0 then
             begin
               InsertString('Constant evaluation:');
               XDecomp;
             end;
    end;
  end;
  end;

  
 procedure TPLFormula.Implication;{converting to >,~}
 var pchar:char;pompt1,pompt2,pompt3:PSTreeNode;
 begin
   if p<>nil then
     begin
       pchar:='>';
       if p^.character = '=' then
         begin
           RCopyTree(p^.left,pompt2);RCopyTree(p^.right,pompt1);
           new(pompt3);ChangeOper(pompt3,pchar);pompt3^.neg:=false;
           pompt3^.right:=p^.right;pompt3^.left:=p^.left;p^.left:=pompt3;
           new(pompt3);ChangeOper(pompt3,pchar);pompt3^.neg:=true;
           pompt3^.left:=pompt1;pompt3^.right:=pompt2;p^.right:=pompt3;
           ChangeOper(p,pchar);p^.neg:=not(p^.neg);
           if Protocol<>nil then
           with Protocol^ do
           if Format^.manner <= 2  then
             begin
               if Format^.manner <= 1 then
                   InsertString('Equivalence to implication:');
               XDecomp;
             end;
         end
         { ( A = B ) = ~ ( ( A > B ) > ~ ( B > A ) ) }
       else if p^.character = '&' then
         begin
           ChangeOper(p,pchar);p^.neg:=not(p^.neg);
           p^.right^.neg:=not(p^.right^.neg);
           if Protocol<>nil then
           with Protocol^ do
           if Format^.manner <= 2  then
             begin
               if Format^.manner <= 1 then
                   InsertString('Conjunction to implication:');
               XDecomp;
             end;
         end
         { ( A & B ) = ~ ( A > ~ B ) }
       else if p^.character = '|' then
         begin
           ChangeOper(p,pchar);p^.left^.neg:=not(p^.left^.neg);
           if Protocol<>nil then
           with Protocol^ do
           if Format^.manner <= 2  then
             begin
               if Format^.manner <= 1 then
                   InsertString('Disjunction to implication:');
               XDecomp;
             end;
         end;
         { ( A | B ) =  ~ A > B  }
     end;
  end;

  procedure TPLFormula.Disjunction;{converting to |,~}
  var pchar:char;pompt1,pompt2,pompt3:PSTreeNode;
  begin
      if p<>nil then
     begin
       pchar:='|';
       if p^.character = '=' then
         begin
           if p^.neg = true then
           begin
             RCopyTree(p^.left,pompt2);RCopyTree(p^.right,pompt1);
             new(pompt3);ChangeOper(pompt3,pchar);pompt3^.neg:=true;
             pompt3^.right:=p^.right;pompt3^.left:=p^.left;
             pompt3^.left^.neg:=not(pompt3^.left^.neg);p^.left:=pompt3;
             new(pompt3);ChangeOper(pompt3,pchar);pompt3^.neg:=true;
             pompt3^.left:=pompt1;pompt3^.right:=pompt2;
             pompt3^.left^.neg:=not(pompt3^.left^.neg);p^.right:=pompt3;
             ChangeOper(p,pchar);p^.neg:=not(p^.neg);

           end
           else
           begin
             RCopyTree(p^.left,pompt2);RCopyTree(p^.right,pompt1);
             new(pompt3);ChangeOper(pompt3,pchar);pompt3^.neg:=true;
             pompt3^.right:=p^.right;pompt3^.left:=p^.left;p^.left:=pompt3;
             new(pompt3);ChangeOper(pompt3,pchar);pompt3^.neg:=true;
             pompt3^.left:=pompt1;pompt3^.right:=pompt2;
             pompt3^.left^.neg:=not(pompt3^.left^.neg);
             pompt3^.right^.neg:=not(pompt3^.right^.neg);p^.right:=pompt3;
             ChangeOper(p,pchar);
           end;
          if Protocol<>nil then
           with Protocol^ do
           if Format^.manner <= 2  then
             begin
               if Format^.manner <= 1 then
                   InsertString('Equivalence to disjunction:');
               XDecomp;
             end;
         end
         { ( A = B ) = ~ ( ~ ( ~ A | B ) | ~ ( B | ~ A ) )
           ( A = B ) =  ~ (  A | B ) | ~ ( ~ B | ~ A )  }
       else if p^.character = '&' then
         begin
           ChangeOper(p,pchar);p^.neg:=not(p^.neg);
           p^.right^.neg:=not(p^.right^.neg);
           p^.left^.neg:=not(p^.left^.neg);
           if Protocol<>nil then
           with Protocol^ do
           if Format^.manner <= 2  then
             begin
               if Format^.manner <= 1 then
                   InsertString('Conjunction to disjunction:');
               XDecomp;
             end;
         end
         { ( A & B ) = ~ ( ~ A | ~ B ) }
       else if p^.character = '>' then
         begin
           ChangeOper(p,pchar);p^.left^.neg:=not(p^.left^.neg);
           if Protocol<>nil then
           with Protocol^ do
           if Format^.manner <= 2  then
             begin
               if Format^.manner <= 1 then
                   InsertString('Implication to disjunction:');
               XDecomp;
             end;
         end;
         { ( A > B ) =  ~ A | B  }
     end;
  end;

  procedure TPLFormula.Conjunction;{converting to &,~}
  var pchar:char;pompt1,pompt2,pompt3:PSTreeNode;
  begin
     if p<>nil then
     begin
       pchar:='&';
       if p^.character = '=' then
         begin
           if p^.neg = true then
           begin
             RCopyTree(p^.left,pompt2);RCopyTree(p^.right,pompt1);
             new(pompt3);ChangeOper(pompt3,pchar);pompt3^.neg:=true;
             pompt3^.right:=p^.right;pompt3^.left:=p^.left;
             pompt3^.left^.neg:=not(pompt3^.left^.neg);
             pompt3^.right^.neg:=not(pompt3^.right^.neg);p^.left:=pompt3;
             new(pompt3);ChangeOper(pompt3,pchar);pompt3^.neg:=true;
             pompt3^.left:=pompt1;pompt3^.right:=pompt2;p^.right:=pompt3;
             ChangeOper(p,pchar);p^.neg:=not(p^.neg);
           end
           else
           begin
             RCopyTree(p^.left,pompt2);RCopyTree(p^.right,pompt1);
             new(pompt3);ChangeOper(pompt3,pchar);pompt3^.neg:=true;
             pompt3^.right:=p^.right;pompt3^.left:=p^.left;
             pompt3^.right^.neg:=not(pompt3^.right^.neg);p^.left:=pompt3;
             new(pompt3);ChangeOper(pompt3,pchar);pompt3^.neg:=true;
             pompt3^.left:=pompt1;pompt3^.right:=pompt2;
             pompt3^.right^.neg:=not(pompt3^.right^.neg);
             p^.right:=pompt3;
             ChangeOper(p,pchar);
           end;
           if Protocol<>nil then
           with Protocol^ do
           if Format^.manner <= 2  then
             begin
               if Format^.manner <= 1 then
                   InsertString('Equivalence to conjunction:');
               XDecomp;
             end;
         end
         { ( A = B ) = ~ ( ~ ( ~ A & ~ B ) & ~ ( B &  A ) )
           ( A = B ) =  ~ (  A & ~ B ) & ~ ( ~ B &  A )  }
       else if p^.character = '|' then
         begin
           ChangeOper(p,pchar);p^.neg:=not(p^.neg);
           p^.right^.neg:=not(p^.right^.neg);
           p^.left^.neg:=not(p^.left^.neg);
           if Protocol<>nil then
           with Protocol^ do
           if Format^.manner <= 2  then
             begin
               if Format^.manner <= 1 then
                   InsertString('Disjunction to conjunction:');
               XDecomp;
             end;
         end
         { ( A | B ) = ~ ( ~ A & ~ B ) }
       else if p^.character = '>' then
         begin
           ChangeOper(p,pchar);p^.neg:=not(p^.neg);
           p^.right^.neg:=not(p^.right^.neg);
           if Protocol<>nil then
           with Protocol^ do
           if Format^.manner <= 2  then
             begin
               if Format^.manner <= 1 then
                   InsertString('Implication to conjunction:');
               XDecomp;
             end;
         end;
         { ( A > B ) =  ~ ( A & ~ B )  }
     end;
  end;

  procedure TPLFormula.Distribution;
  { ( A & B ) | C = ( A | C ) & ( B | C )
    ( A | B ) & C = ( A & C ) | ( B & C ) }
    var pchar,pchar2:char;pompt1,pompt2,pompt3:PSTreeNode;
  begin
    if (p1<>nil) and (p2<>nil) and (p1^.character in ['&','|'] )
      and (p2^.character in ['&','|']) then
      begin
        RCopyTree(p2^.right,pompt1);
        pchar:=p1^.character;
        pchar2:=p2^.character;
        Changeoper(p2,pchar);Changeoper(p1,pchar2);pompt2:=p1^.right;
        p1^.right:=pompt1;new(pompt3);Changeoper(pompt3,pchar2);
        pompt3^.neg:=false;pompt3^.left:=pompt2;
        pompt3^.right:=p2^.right;p2^.right:=pompt3;
        if Protocol<>nil then
           with Protocol^ do
           if Format^.manner <= 2  then
             begin
               if Format^.manner <= 1 then
                   InsertString('Distribution:');
               XDecomp;
             end;
      end;
   end;

   procedure TPLFormula.IdCom;
   begin
     if (p1<>nil)and(p2<>nil)and(p1^.prior=6)and(p2^.prior=6)and
        (not(p1^.character in ['0','1']))and
        (not(p2^.character in ['0','1'])) then
       begin
         if p1^.character=p2^.character then
           begin
             if p1^.neg=p2^.neg then
               begin
                 case master^.character of
                 '&':p1^.character:='1';
                 '|':p1^.character:='0';
                 end;
                 p1^.neg:=false;
                 if Protocol<>nil then
                 with Protocol^ do
                  if Format^.manner <= 2  then
                   begin
                    if Format^.manner <= 1 then
                     InsertString('Idempotention:');
                    XDecomp;
                   end;
               end
             else
             begin
                 case master^.character of
                  '&':p1^.character:='0';
                  '|':p1^.character:='1';
                  end;
                 p1^.neg:=false;
                 if Protocol<>nil then
                 with Protocol^ do
                 if Format^.manner <= 2  then
                  begin
                    if Format^.manner <= 1 then
                     InsertString('Complementarity:');
                    XDecomp;
                 end;
             end;
           end;
       end;
   end;

   procedure TPLFormula.AbsExt;
   var OK,left,right,cst:boolean;pom1,pom2,pom3,pom4,pt:PSTreeNode;
     procedure Find(var tr:PSTreeNode);
     begin
       if (tr^.prior<>6) then
         begin
           Find(tr^.left);
           Find(tr^.right);
         end
       else begin
              if not(tr^.character in ['0','1']) then 
              if pt=nil then pt:=tr;
              if tr^.character in ['0','1'] then cst:=true;
            end;
     end;

   { absorption,negation absorption and extension law
     on NF elements }
   begin
     pom1:=p1;pom2:=p2;pom3:=nil;pom4:=nil;Ok:=true;left:=false;
     right:=false;cst:=false;
     pt:=nil;Find(p1);pom1:=pt;
     if cst=false then
       begin
         pt:=nil;Find(p2);pom2:=pt;
       end;
     if (pom1<>nil)and(pom2<>nil)and(cst=false) then
     begin
     repeat

       if ((pom1<>nil)and(pom2<>nil)and(pom1^.character>pom2^.character))or
          ((pom1=nil)and(pom2<>nil)) then
         begin
           right:=true;
           if pom2<>nil then pom2:=pom2^.right;
				 end;
			 if ((pom1<>nil)and(pom2<>nil)and(pom1^.character<pom2^.character))or
					((pom1<>nil)and(pom2=nil)) then
				 begin
					 left:=true;
					 if pom1<>nil then pom1:=pom1^.right;
				 end;
			 if (pom1^.character=pom2^.character)and(pom1^.neg<>pom2^.neg) then
				 if (pom3<>nil)and(pom4<>nil) then OK:=false
				 else
					 begin
						 pom3:=pom1;
						 pom4:=pom2;
					 end;
			 if (pom1<>nil)and(pom2<>nil)and(pom1^.character=pom2^.character) then
					 begin
						 if (pom1<>nil) then pom1:=pom1^.right;
						 if (pom2<>nil) then pom2:=pom2^.right;
					 end;
			 until (OK=false)or((pom1=nil)and(pom2=nil));
			 pt:=nil;Find(p1);pom1:=pt;
			 pt:=nil;Find(p2);pom2:=pt;
       if (not((left)and(right)))and(OK) then
         begin
           if (left=false)and(right=false) then
             if pom3=nil then       {two same elms}
              begin
                case pchar of
                 '&':pom1^.character:='1';
                 '|':pom1^.character:='0';
                end;
                pom1^.neg:=false;
                if Protocol<>nil then
                with Protocol^ do
                 if Format^.manner <= 2  then
                  begin
                    if Format^.manner <= 1 then
                     InsertString('Total absorption:');
                    XDecomp;
                  end;
              end
           else                   {extension law}
             begin
               case pchar of
               '&': begin
                      pom1^.character:='1';
                      pom4^.character:='0';
                    end;
               '|': begin
                      pom1^.character:='0';
                      pom4^.character:='1';
                   end;
               end;
               pom1^.neg:=false;pom4^.neg:=false;
               if Protocol<>nil then
                with Protocol^ do
                 if Format^.manner <= 2  then
                  begin
                    if Format^.manner <= 1 then
                     InsertString('Extension:');
                    XDecomp;
                  end;
             end;
         if left then
           if pom3=nil then
             begin
               case pchar of                {absorption}
               '&': pom1^.character:='1';
               '|': pom1^.character:='0';
               end;
               pom1^.neg:=false;
               if Protocol<>nil then
                with Protocol^ do
                 if Format^.manner <= 2  then
                  begin
                    if Format^.manner <= 1 then
                     InsertString('Absorption:');
                    XDecomp;
                  end;
             end
           else
             begin
               case pchar of                {negation absorption}
               '&': pom3^.character:='0';
               '|': pom3^.character:='1';
               end;
               pom3^.neg:=false;
               if Protocol<>nil then
                with Protocol^ do
                 if Format^.manner <= 2  then
                  begin
                    if Format^.manner <= 1 then
                     InsertString('Absorption of negation:');
                    XDecomp;
                  end;
             end;
         if right then
           if pom4=nil then
             begin
               case pchar of                {absorption}
               '&': pom2^.character:='1';
               '|': pom2^.character:='0';
               end;
               pom2^.neg:=false;
               if Protocol<>nil then
                with Protocol^ do
                 if Format^.manner <= 2  then
                  begin
                    if Format^.manner <= 1 then
                     InsertString('Absorption:');
                    XDecomp;
                  end;
             end
           else
             begin
               case pchar of                {negation absorption}
               '&': pom4^.character:='0';
               '|': pom4^.character:='1';
               end;
               pom4^.neg:=false;
               if Protocol<>nil then
                with Protocol^ do
                 if Format^.manner <= 2  then
                  begin
                    if Format^.manner <= 1 then
                     InsertString('Absorption of negation:');
                    XDecomp;
                  end;
             end;
         Thread;
         o:=true;
     end;
   end;
   end;


  {user procedures for manipulations}

  procedure TPLFormula.RConstEval;
    begin
      if  (tr^.prior <> 6) then
        begin
          RConstEval(tr^.left);
          RConstEval(tr^.right);
          EvlConst1(tr);
        end;
    end;
  procedure TPLFormula.ConstEval;
  begin
    if Stree<>nil then
      begin
        RConstEval(Stree);
        
      end;  
  end;


  procedure TPLFormula.RToImp;
  begin
    if  (tr^.prior <> 6) then
      begin
        Implication(tr);
        RToImp(tr^.left);
        RToImp(tr^.right);
      end;
  end;

  procedure TPLFormula.ToImp;
  begin
    if Stree<>nil then
      begin
        RToImp(Stree);
        if Protocol<>nil then
                with Protocol^ do
                 if Format^.manner = 3  then
                  begin
                    InsertString('Result - Global conversion to implication:');
                    XDecomp;
                  end;
      end;
  end;

  procedure TPLFormula.RToDis;
  begin
    if (tr^.prior <> 6) then
      begin
        Disjunction(tr);
        RToDis(tr^.left);
        RToDis(tr^.right);
      end;
  end;

  procedure TPLFormula.ToDis;
  begin
    if Stree<>nil then
      begin
        RToDis(Stree);
        if Protocol<>nil then
                with Protocol^ do
                 if Format^.manner = 3  then
                  begin
                    InsertString('Result - Global conversion to disjunctions:');
                    XDecomp;
                  end;
      end;
  end;

  procedure TPLFormula.RToCon;
  begin
    if (tr^.prior<>6) then
      begin
        Conjunction(tr);
        RToCon(tr^.left);
        RToCon(tr^.right);
      end;
  end;

  procedure TPLFormula.ToCon;
  begin
    if Stree<>nil then
      begin
        RToCon(Stree);
        if Protocol<>nil then
                with Protocol^ do
                 if Format^.manner = 3  then
                  begin
                    InsertString('Result - Global conversion to conjunctions:');
                    XDecomp;
                  end;
      end;
  end;

  procedure TPLFormula.RToConDis;
  begin
    if (tr^.prior<>6) then
      begin
        if (tr^.neg)or(not(tr^.character in ['&','|'])) then Conjunction(tr);
        if tr^.neg then Disjunction(tr);
        RToConDis(tr^.left);
        RToConDis(tr^.right);
      end;
  end;

  procedure TPLFormula.ToConDis;
  begin
    if Stree<>nil then
     begin
       RToConDis(Stree);
       if Protocol<>nil then
                with Protocol^ do
                 if Format^.manner = 3  then
                  begin
                    InsertString('Result - Global conversion to literal form:');
                    XDecomp;
                  end;
     end;
  end;

  procedure TPLFormula.ReduceElm;
  var pompt1:PSTreeNode;
  procedure RReduceElm(var tr:PSTreeNode);
  
  begin
  if (tr^.character=t^.character)or(tr^.prior=6) then
  begin
    if (tr^.prior<>6) then
      begin
        RReduceElm(tr^.left);
        RReduceElm(tr^.right);
      end
    else 
      begin
        if (pompt1<>nil) then IdCom(pompt1,tr,t);
        pompt1:=tr;
      end;
  end;
  end;
  begin
    pompt1:=nil;
    RReduceElm(t);
  end;

  procedure TPLFormula.Exchange;
  var pom:PSTreeNode;
  begin
    new(pom);
    ShCopyNode(p1,pom);
    ShCopyNode(p2,p1);
    ShCopyNode(pom,p2);
    dispose(pom);
  end;


  procedure TPLFormula.SortElm;
    var op:boolean;var pchar:char;pompt1:PSTreeNode;
    procedure Bubble(var t:PSTreeNode);
      begin
        if (t^.character=pchar)or(t^.prior=6) then
        begin
          if t^.prior<>6 then
            begin
              Bubble(t^.left);
              Bubble(t^.right);
            end
          else
          begin
            if (pompt1<>nil)and(pompt1^.character>t^.character) then
              begin
                Exchange(pompt1,t);
                op:=true;
              end;
            pompt1:=t;
          end;
        end;
      end;

    begin
      op:=true;
      pchar:=tr^.character;
      while op do
       begin
        op:=false;
        pompt1:=nil;
        Bubble(tr);
       end;
    end;

  procedure TPLFormula.RToCNF;
  begin
    if  (tr^.prior<>6)and(not(LowMemory)) then
      begin
        if tr^.character<>andchar then
          begin
            FindOper(tr,andchar);
            if tr^.left^.character<>andchar then
               begin
                 SortElm(tr);
                 ReduceElm(tr);
               end;
            if tr^.left^.character=andchar then Distribution(tr^.left,tr);
          end;
        RToCNF(tr^.left);
        RToCNF(tr^.right);
        if (tr^.character=andchar) then
         begin
           RConstEval(tr^.left);
           RConstEval(tr^.right);
         end;
      end;

  end;

  
  procedure TPLFormula.ToCNF;
  var i:integer;
  begin
    if Stree<>nil then
    begin
      ToConDis;
      ConstEval;
      RToCNF(Stree);
      if LowMemory then
        begin
          i:=DoEditDialog(edOutOfMemory,nil);
          DisposeTree;
          Stree:=nil;
        end
      else
      begin
        SortElm(Stree);
        ReduceElm(STree);
        ConstEval;
        if Protocol<>nil then
                with Protocol^ do
                 if Format^.manner = 3  then
                  begin
                    InsertString('Result - Conversion to CNF:');
                    XDecomp;
                  end;
       end;
    end;
  end;

  procedure TPLFormula.RToDNF;
  begin
    if (tr^.prior<>6)and(not(LowMemory)) then
      begin
        if tr^.character<>orchar then
          begin
            FindOper(tr,orchar);
            if tr^.left^.character<>orchar then
               begin
                 SortElm(tr);
                 ReduceElm(tr);
               end;
            if tr^.left^.character=orchar then Distribution(tr^.left,tr);
            
          end;
        RToDNF(tr^.left);
        RToDNF(tr^.right);
        if (tr^.character=orchar) then
         begin
           RConstEval(tr^.left);
           RConstEval(tr^.right);
         end;
      end;
  end;

  procedure TPLFormula.ToDNF;
  var i:integer;
  begin
    if Stree<>nil then
     begin
      ToConDis;
      ConstEval;
      RToDNF(Stree);
      if LowMemory then
        begin
          i:=DoEditDialog(edOutOfMemory,nil);
          DisposeTree;
          Stree:=nil;
        end
      else
      begin
        SortElm(Stree);
        ReduceElm(STree);
        ConstEval;
        if Protocol<>nil then
                with Protocol^ do
                 if Format^.manner = 3  then
                  begin
                    InsertString('Result - Conversion to DNF:');
                    XDecomp;
                  end;
       end;
     end;
  end;

  procedure TPLFormula.Thread;
  var pompt1:PSTreeNode;pchar:char;
    procedure RThread(var tr:PSTreeNode);
    begin
      if tr^.prior<>6 then
        begin
          RThread(tr^.left);
          RThread(tr^.right);
        end
      else if not(tr^.character in ['1','0']) then
        begin
          if pompt1<>nil then
            begin
              pompt1^.right:=tr;tr^.left:=pompt1;
            end;
          pompt1:=tr;
        end;
    end;

    procedure RDive(var t:PSTreeNode);
    begin
      if (t^.prior<>6) then
        if t^.character=pchar then
          begin
            RDive(t^.left);
            RDive(t^.right);
          end
        else if not(t^.character in ['0','1']) then
        begin
          pompt1:=nil;
          RThread(t);
        end;
    end;
    
  begin
    pompt1:=nil;pchar:=Stree^.character;RDive(Stree);
  end;

  procedure TPLFormula.UnThread;
    procedure RUnThread(var tr:PSTreeNode);
    begin
      if tr^.prior<>6 then
        begin
          RUnThread(tr^.left);
          RUnThread(tr^.right);
        end
      else
        begin
          tr^.right:=nil;tr^.left:=nil;
        end;
    end;
  begin
    RUnThread(STree);
  end;

  procedure TPLFormula.Minimize;
  var pompt1,pompt2:PSTreeNode;pchar:char;op:boolean;

    procedure RRunner(var t:PSTreeNode);
      begin
        if (t^.prior<>6)and(t^.character=pchar) then
          begin
            RRunner(t^.left);
            RRunner(t^.right);
          end;
        if (t^.character<>pchar)and(not(t^.character in ['0','1'])) then
        begin
         if pompt1=t then pompt1:=nil;
         if pompt1<>nil then
          begin
            AbsExt(t,pompt1,op,pchar);

            
          end;
        end;
      end;

    procedure RMinimize(var tr:PSTreeNode);
      begin
        if (tr^.prior<>6)and(tr^.character=pchar) then
          begin
            RMinimize(tr^.left);
            RMinimize(tr^.right);
          end;
        if (tr^.character<>pchar)and(not(tr^.character in ['0','1'])) then
          begin
            pompt1:=tr;RRunner(STree);
          end;
      end;

   begin
     if (STree<>nil)and(Stree^.prior<>6) then
       begin
         op:=true;
         Thread;
         while op do
           begin
             pchar:=STree^.character;
             op:=false;
             pompt1:=nil;
             RMinimize(STree);
             ConstEval;
             UnThread;
             Thread;
           end;
         UnThread;
         if STree^.prior<>6 then
               begin
                 SortElm(STree);
                 ReduceElm(STree);
                 ConstEval;
               end;

         if Protocol<>nil then
                with Protocol^ do
                 if Format^.manner = 3  then
                  begin
                    InsertString('Result - Reduction to minimal NF:');
                    XDecomp;
                  end;
       end;
    end;



  {Turbo Vision Declarations}

  constructor TPLFormula.Init;
  begin
    Stree:=nil;
    Protocol:=nil;
  end;

  destructor TPLFormula.Done;
  begin
    DisposeTree;
  end;

  constructor TProtocol.Init;
  begin
    TEditWindow.Init(Bounds,
      FileName, ANumber);
    Formula:=new(PPLFormula,Init);
    Formula^.Protocol:=@self;
    Format:=new(PFormat);
    Format^.line:=77;
    Format^.space:=false;
    Format^.results:=true;
    Format^.manner:=4;
  end;

  destructor TProtocol.Done;
  begin
    dispose(Formula,done);
    dispose(Format);
    TEditWindow.Done;
  end;

 procedure TProtocol.HandleEvent;
 var i: integer;                              
 begin
   if (GetState(sfFocused)) then EnableCommands([cmEval,cmFormat,
    cmMinimalCNF,cmMinimalDNF,cmCNF,cmDecision,cmDNF,cmImp,cmDecompile,
    cmCompile,cmConDis,cmCon,cmDis]);
   if (Event.What = evCommand)and(Event.Command in [cmClose,cmNext,cmPrev])
     then DisableCommands([cmEval,cmFormat,
    cmMinimalCNF,cmMinimalDNF,cmCNF,cmDecision,cmDNF,cmImp,cmDecompile,
    cmCompile,cmConDis,cmCon,cmDis]);    
   TEditWindow.HandleEvent(Event);
   case Event.What of
    evCommand:
    begin
      if (Editor^.SelStart <> Editor^.SelEnd)and
         (Event.Command in [cmDis,cmCon,cmConDis,cmImp,cmDNF,
          cmCNF,cmEval,cmMinimalCNF,cmMinimalDNF,cmDecision])
          then
           begin
            XComp;
            if Error <> OK then ClearEvent(Event);
           end;
      if Event.What=evCommand then
      case Event.Command of
        cmCompile: begin
                    if Editor^.SelStart<Editor^.SelEnd then XComp
                     else i:=DoEditDialog(edNothing,nil);
                   end;
        cmDecompile: begin
                      XDecomp;
                     end;
        cmDis      : begin
                       Formula^.ToDis;
                       if Format^.results then XDecomp;
                     end;
        cmCon      : begin
                       Formula^.ToCon;
                       if Format^.results then XDecomp;
                     end;
        cmConDis   : begin
                       Formula^.ToConDis;
                       if Format^.results then XDecomp;
                     end;
        cmImp      : begin
                       Formula^.ToImp;
                       if Format^.results then XDecomp;
                     end;
        cmDNF      : begin
                       Formula^.ToDNF;
                       if Format^.results then XDecomp;
                     end;
        cmCNF      : begin
                       Formula^.ToCNF;
                       if Format^.results then XDecomp;
                     end;
        cmEval     : begin
                       Formula^.ConstEval;
                       if Format^.results then XDecomp;
                     end;
        cmMinimalDNF :begin
                         Formula^.ToDNF;
                         Formula^.Minimize;
                         if Format^.results then XDecomp;
                       end;
        cmMinimalCNF :begin
                         Formula^.ToCNF;
                         Formula^.Minimize;
                         if Format^.results then XDecomp;
                      end;
        cmDecision   :begin
                        Decision;
                      end;  
        cmFormat : begin
                     DoDialog;
                   end;
        
      else Exit;
      end;
   end;
   end;
   ClearEvent(Event);
 end; 

 


constructor TEditorApp.Init;
var
  R: TRect;i:integer;
begin
  LowMemSize:=Heapsize;
  MaxHeapSize:=MemAvail div 32;
  TApplication.Init;
  DisableCommands([cmSave, cmSaveAs, cmCut, cmCopy, cmPaste, cmClear,
    cmUndo, cmFind, cmReplace, cmSearchAgain, cmEval,cmFormat,
    cmMinimalCNF,cmMinimalDNF,cmCNF,cmDecision,cmDNF,cmImp,cmDecompile,
    cmCompile,cmConDis,cmCon,cmDis]);
  EditorDialog := DoEditDialog;
  ClipWindow := OpenEditor('', False);
  if ClipWindow <> nil then
  begin
    Clipboard := ClipWindow^.Editor;
    Clipboard^.CanUndo := False;
  end;
  RegisterHelpFile;
  MessageBox('(c) Hashim Habiballa, 1996.',
  nil,mfInformation+mfOKButton);
end;

function TEditorApp.GetPalette;
const
  CNewColor = CColor + CHelpColor;
  CNewBlackWhite = CBlackWhite + CHelpBlackWhite;
  CNewMonochrome = CMonochrome + CHelpMonochrome;
  P: array[apColor..apMonochrome]
     of string[length(CNewColor)] =(CNewColor,CNewBlackWhite,CNewMonochrome);
begin
  GetPalette:=@P[AppPalette];
end;

procedure TEditorApp.GetEvent;
var W:PWindow;HFile:PHelpFile;
    HelpStrm: PDosStream;
    Pom:TPoint;
const HelpInUse: Boolean = false;
begin
  TApplication.GetEvent(Event);
  if (not(HelpInUse))and(Event.What=evCommand)
    and(Event.Command in [cmHelp,cmAbout,cmSystem,cmPL,cmContents]) then
       begin
         HelpInUse:=True;
         HelpStrm:=new(PDosStream,Init('deal.hlp',stOpenRead));
         HFile:=new(PHelpFile,Init(HelpStrm));
         if HelpStrm^.Status <> stOK then
           begin
             MessageBox('Help file can`t be opened.',nil,mfError+mfOKButton);
             Dispose(HFile,Done);
           end
         else
           begin
             case Event.Command of
             cmHelp:W:=New(PHelpWindow,Init(HFile,GetHelpCtx));
             cmPL:W:=New(PHelpWindow,Init(HFile,hcPL));
             cmAbout:W:=New(PHelpWindow,Init(HFile,hcAbout));
             cmSystem:W:=New(PHelpWindow,Init(HFile,hcSystem));
             cmContents:W:=New(PHelpWindow,Init(HFile,hcContents));
             end;
             if ValidView(W)<>nil then
               begin
                 ExecView(W);
                 Dispose(W,Done);
               end;
            ClearEvent(Event);
            end;
        HelpInUse:=False;
       end;
end;

procedure TEditorApp.HandleEvent(var Event: TEvent);
const n:integer=0;
procedure FileOpen;
var
  FileName: FNameStr;
begin
  FileName := '*.*';
  if ExecDialog(New(PFileDialog, Init('*.*', 'Open file',
    '~N~ame', fdOpenButton, 100)), @FileName) <> cmCancel then
    OpenEditor(FileName, True);
end;

procedure ProtocolOpen;
var
  FileName: FNameStr;
begin
  FileName := '*.*';
  if ExecDialog(New(PFileDialog, Init('*.*', 'Open protocol',
    '~N~ame', fdOpenButton, 100)), @FileName) <> cmCancel then
    OpenProtocol(FileName, True);
end;




procedure FileNew;
begin
  OpenEditor('', True);
end;

procedure ProtocolNew;
begin
  OpenProtocol('', True);

end;


procedure ChangeDir;
begin
  ExecDialog(New(PChDirDialog, Init(cdNormal, 0)), nil);
end;

procedure ShowClip;
begin
  ClipWindow^.Select;
  ClipWindow^.Show;
end;


begin
  TApplication.HandleEvent(Event);
  case Event.What of
    evCommand:
    begin
      case Event.Command of
        cmOpen: FileOpen;
        cmNew: FileNew;
        cmChangeDir: ChangeDir;
        cmDosShell: DosShell;
        cmShowClip: ShowClip;
        cmTile: Tile;
        cmCascade: Cascade;
        cmNewProt: ProtocolNew;
        cmRetrieve :ProtocolOpen;

      else
        Exit;
      end;
    end;
  else
    Exit;
  end;
  ClearEvent(Event);
end;

procedure TEditorApp.InitMenuBar;
var
  R: TRect;
begin
  GetExtent(R);
  R.B.Y := R.A.Y + 1;
  MenuBar := New(PMenuBar, Init(R, NewMenu(
    NewSubMenu('~F~ile', hcFile, NewMenu(
      NewItem('~O~pen...', 'F3', kbF3, cmOpen, hcOpen,
      NewItem('~N~ew', '', kbNoKey, cmNew, hcNew,
      NewItem('~S~ave', 'F2', kbF2, cmSave, hcSave,
      NewItem('S~a~ve as...', '', kbNoKey, cmSaveAs, hcSaveas,
      NewLine(
      NewItem('~C~hange dir...', '', kbNoKey, cmChangeDir, hcChangeDir,
      NewItem('~D~OS shell', '', kbNoKey, cmDosShell, hcDosShell,
      NewItem('E~x~it', 'Alt-X', kbAltX, cmQuit, hcExit,
      nil))))))))),
    NewSubMenu('~E~dit', hcEdit, NewMenu(
      NewItem('~U~ndo', '', kbNoKey, cmUndo, hcUndo,
      NewLine(
      NewItem('Cu~t~', 'Shift-Del', kbShiftDel, cmCut, hcCut,
      NewItem('~C~opy', 'Ctrl-Ins', kbCtrlIns, cmCopy, hcCopy,
      NewItem('~P~aste', 'Shift-Ins', kbShiftIns, cmPaste, hcPaste,
      NewItem('~S~how clipboard', '', kbNoKey, cmShowClip, hcShowClip,
      NewLine(
      NewItem('~C~lear', 'Ctrl-Del', kbCtrlDel, cmClear, hcClear,
      nil))))))))),
    
    NewSubMenu('~W~indows', hcWindows, NewMenu(
      NewItem('~S~ize/move','Ctrl-F5', kbCtrlF5, cmResize, hcSizeMove,
      NewItem('~Z~oom', 'F5', kbF5, cmZoom, hcZoom,
      NewItem('~T~ile', '', kbNoKey, cmTile, hcTile,
      NewItem('C~a~scade', '', kbNoKey, cmCascade, hcCascade,
      NewItem('~N~ext', 'F6', kbF6, cmNext, hcNext,
      NewItem('~P~revious', 'Shift-F6', kbShiftF6, cmPrev, hcPrevious,
      NewItem('~C~lose', 'Alt-F3', kbAltF3, cmClose, hcClose,
      nil)))))))),
    NewSubMenu('~P~rotocol',hcProtocol, NewMenu(
      NewItem('~F~ormat...','Alt-T', kbAltT,cmFormat, hcFormat,
      NewItem('~R~etrieve file','Ctrl-F3',kbCtrlF3,cmRetrieve, hcRetrieve,
      NewItem('~N~ew', 'Alt-N' ,kbAltN,cmNewProt,hcPNew,
      nil)))),
    NewSubMenu('Fo~r~mula',hcFormula,NewMenu(
      NewSubMenu('~C~onversion', hcConversion,NewMenu(
        NewItem('~C~onjunctions','Alt-C',kbAltC,cmCon,hcConjunctions,
        NewItem('~D~isjunctions','Alt-D',kbAltD,cmDis,hcDisjunctions,
        NewItem('~I~mplications','Alt-I',kbAltI,cmImp,hcImplications,
        NewItem('~L~iterals (&,|)','Alt-L',kbAltL,cmConDis,hcLiterals,
        nil))))),
      NewItem('Co~m~pile','Ctrl-F9',kbCtrlF9,cmCompile, hcCompile,
      NewItem('~D~ecompile', 'Shift-F9',kbShiftF9,cmDecompile,hcDecompile,
      NewSubMenu('~N~ormal forms',hcNormalForms,NewMenu(
        NewItem('~D~isjunctive','Alt-F7',kbAltF7,cmDNF,hcDisjunctive,
        NewItem('~C~onjuctive','Ctrl-F7',kbCtrlF7,cmCNF,hcConjunctive,
        NewItem('~M~inimal DNF','Alt-F8',kbAltF8,cmMinimalDNF,hcMinimalDNF,
        NewItem('M~i~nimal CNF','Ctrl-F8',kbCtrlF8,cmMinimalCNF,hcMinimalCNF,
        nil))))),
      NewItem('~E~valuate', 'Alt-V',kbAltV,cmEval,hcEvaluate,
      NewItem('Dec~i~sion', 'F8',kbF8,cmDecision,hcMDecision,
      nil))))))),
    NewSubMenu('~H~elp',hcNoContext,NewMenu(
      NewItem('~C~ontents','Alt-F1',kbAltF1,cmContents,hcNoContext,
      NewItem('~A~bout help','',kbNoKey,cmAbout,hcNoContext,
      NewItem('A~b~out system','',kbNoKey,cmSystem,hcNoContext,
      NewItem('~P~ropositional logic','',kbNoKey,cmPL,hcNoContext,
      nil))))),
    nil)))))))));
end;

procedure TEditorApp.InitStatusLine;
var
  R: TRect;
begin
  GetExtent(R);
  R.A.Y := R.B.Y - 1;
  New(StatusLine, Init(R,
    NewStatusDef(0, $FFFF,
      NewStatusKey('~F1~ Help',kbF1,cmHelp,
      NewStatusKey('~F2~ Save', kbF2, cmSave,
      NewStatusKey('~F3~ Open', kbF3, cmOpen,
      NewStatusKey('~Alt-F3~ Close', kbAltF3, cmClose,
      NewStatusKey('~F5~ Zoom', kbF5, cmZoom,
      NewStatusKey('~F6~ Next', kbF6, cmNext,
      NewStatusKey('~F10~ Menu', kbF10, cmMenu,
      NewStatusKey('', kbCtrlF5, cmResize,
      nil)))))))),
    nil)));
end;

procedure TEditorApp.OutOfMemory;
begin
  MessageBox('Not enough memory for this operation.',
    nil, mfError + mfOkButton);
end;

begin
  EditorApp.Init;
  EditorApp.Run;
  EditorApp.Done;
end.
